/*
 * Copyright (C) 2020 Activision Publishing, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * 3. Neither the name of Activision Publishing, Inc. nor the names of
 *    its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "RemoteInspectorSocket.h"

#if ENABLE(REMOTE_INSPECTOR)

namespace Inspector {

namespace Socket {

using namespace Telescope;

void init()
{
}

Optional<PlatformSocketType> connect(const char *serverAddress, uint16_t serverPort)
{
	SockAddr_In address;
	
	address.sin_family = TEL_NET_AF_INET;
	inet_aton(serverAddress, &address.sin_addr);
	address.sin_port = htons(serverPort);

	TEL_NET_SOCKET fd = socket(TEL_NET_AF_INET, TEL_NET_SOCK_STREAM, TEL_NET_IPPROTO_TCP);
	if (fd < 0)
	{
		LOG_ERROR("Failed to create socket for %s:%d, errno=%d", serverAddress, serverPort, errno);
		return WTF::nullopt;
	}

	int error = connect(fd, reinterpret_cast<SockAddr*>(&address), sizeof(address));
	if (error < 0)
	{
		LOG_ERROR("Fail to connect to %s:%u, errno=%d", serverAddress, serverPort, errno);
		Telescope::close(fd);
		return WTF::nullopt;
	}

	return fd;
}


Optional<PlatformSocketType> listen(const char *addressStr, uint16_t port)
{
	SockAddr_In address;

	TEL_NET_SOCKET fdListen = socket(TEL_NET_AF_INET, TEL_NET_SOCK_STREAM, TEL_NET_IPPROTO_TCP);
	if (fdListen < 0)
	{
		LOG_ERROR("socket() failed, errno=%d", errno);
		return WTF::nullopt;
	}

	const int enabled = 1;
	int error = setsockopt(fdListen, TEL_NET_SOL_SOCKET, TEL_NET_SO_REUSEADDR, &enabled, sizeof(enabled));
	if (error < 0)
	{
		LOG_ERROR("setsockopt() TEL_NET_SO_REUSEADDR, errno=%d", errno );
		Telescope::close(fdListen);
		return WTF::nullopt;
	}
	error = setsockopt(fdListen, TEL_NET_SOL_SOCKET, TEL_NET_SO_REUSEPORT, &enabled, sizeof(enabled));
	if (error < 0)
	{
		LOG_ERROR("setsockopt() TEL_NET_SO_REUSEPORT, errno=%d", errno );
		Telescope::close(fdListen);
		return WTF::nullopt;
	}

	// FIXME: Support AF_INET6 connections.
	address.sin_family = TEL_NET_AF_INET;
	if (addressStr && *addressStr)
	{
		inet_aton(addressStr, &address.sin_addr);
	}
	else
	{
		address.sin_addr = htonl(TEL_NET_INADDR_ANY);
	}
	address.sin_port = htons(port);
	error = Telescope::bind(fdListen, reinterpret_cast<SockAddr*>(&address), sizeof(address));
	if (error < 0)
	{
		LOG_ERROR("bind() failed, errno = %d", errno);
		Telescope::close(fdListen);
		return WTF::nullopt;
	}

	error = Telescope::listen(fdListen, 1);
	if (error < 0) {
        LOG_ERROR("listen() failed, errno = %d", errno);
        Telescope::close(fdListen);
        return WTF::nullopt;
	}

	return fdListen;
}


Optional<PlatformSocketType> accept(PlatformSocketType socket)
{
	SockAddr_In address;
    size_t len = sizeof(SockAddr_In);
    int fd = Telescope::accept(socket, reinterpret_cast<SockAddr*>(&address), &len);
    if (fd >= 0)
	{
        return fd;
	}

    LOG_ERROR("accept(inet) error (errno = %d)", errno);
    return WTF::nullopt;
}


Optional<std::array<PlatformSocketType, 2>> createPair()
{
    std::array<PlatformSocketType, 2> sockets;

    if (socketepair(TEL_NET_AF_INET, TEL_NET_SOCK_STREAM, 0, &sockets[0]))
	{
        return WTF::nullopt;
	}

    return sockets;
}


bool setup(PlatformSocketType socket)
{
	int nonBlocking = 1;
    if (setsockopt(socket, TEL_NET_SOL_SOCKET, TEL_NET_SO_NBIO, &nonBlocking, sizeof(nonBlocking)))
	{
        LOG_ERROR("setNonBlock() error (errno = %d)", errno);
        return false;
    }

    if (setsockopt(socket, TEL_NET_SOL_SOCKET, TEL_NET_SO_RCVBUF, &BufferSize, sizeof(BufferSize))) 
	{
        LOG_ERROR("setsockopt(TEL_NET_SO_RCVBUF) error (errno = %d)", errno);
        return false;
    }

    if (setsockopt(socket, TEL_NET_SOL_SOCKET, TEL_NET_SO_SNDBUF, &BufferSize, sizeof(BufferSize))) {
        LOG_ERROR("setsockopt(TEL_NET_SO_SNDBUF) error (errno = %d)", errno);
        return false;
    }

    return true;
}


bool isValid(PlatformSocketType socket)
{
	return socket != INVALID_SOCKET_VALUE;
}


Optional<uint16_t> getPort(PlatformSocketType socket)
{
    ASSERT(isValid(socket));

    SockAddr_In address;
    size_t len = sizeof(address);
    if (getsockname(socket, reinterpret_cast<SockAddr*>(&address), &len)) 
	{
        LOG_ERROR("getsockname() error (errno = %d)", errno);
        return WTF::nullopt;
    }
    return address.sin_port;
}


Optional<size_t> read(PlatformSocketType socket, void *buffer, int bufferSize)
{
    ASSERT(isValid(socket));

    int readSize = Telescope::read(socket, buffer, bufferSize);
    if (readSize >= 0)
	{
        return static_cast<size_t>(readSize);
	}

    LOG_ERROR("read error (errno = %d)", errno);
    return WTF::nullopt;
}


Optional<size_t> write(PlatformSocketType socket, const void *data, int size)
{
    ASSERT(isValid(socket));

    int writeSize = Telescope::write(socket, data, size);
    if (writeSize >= 0)
	{
        return static_cast<size_t>(writeSize);
	}

    LOG_ERROR("write error (errno = %d)", errno);
    return WTF::nullopt;
}


void close(PlatformSocketType& socket)
{
    if (!isValid(socket))
	{
        return;
	}

    Telescope::close(socket);
    socket = INVALID_SOCKET_VALUE;
}


PollingDescriptor preparePolling(PlatformSocketType socket)
{
    PollingDescriptor poll = { };
    poll.fd = socket;
    poll.events = TEL_NET_POLLIN;
    return poll;
}


bool poll(Vector<PollingDescriptor> &pollDescriptors, int timeout)
{
    int ret = Telescope::poll(pollDescriptors.data(), pollDescriptors.size(), timeout);
    return ret > 0;
}


bool isReadable(const PollingDescriptor& poll)
{
    return poll.revents & TEL_NET_POLLIN;
}

bool isWritable(const PollingDescriptor& poll)
{
    return poll.revents & TEL_NET_POLLOUT;
}

void markWaitingWritable(PollingDescriptor& poll)
{
    poll.events |= TEL_NET_POLLOUT;
}

void clearWaitingWritable(PollingDescriptor& poll)
{
    poll.events &= ~TEL_NET_POLLOUT;
}


}

}
#endif
