/*
 * Copyright (C) 2020 Activision Publishing, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * 3. Neither the name of Activision Publishing, Inc. nor the names of
 *    its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if USE( TSRENDER )
#include "config.h"
#include "GraphicsContextImpl.h"
#include "GraphicsContextPlatformPrivateTSRender.h"
#include "ShadowBlur.h"
#include "BitmapImage.h"
#include "TSRender.h"
#include "NotImplemented.h"

namespace WebCore
{
void GraphicsContext::platformInit( PlatformGraphicsContext *platformContext )
{
	if ( !platformContext )
	{
		return;
	}

	if ( !m_data )
	{
		m_data = new GraphicsContextPlatformPrivate( platformContext );
		platformContext->setGraphicsContextPrivate( m_data );
	}
}


void GraphicsContext::platformDestroy()
{
	if ( m_data )
	{
		m_data->platformContext->setGraphicsContextPrivate( nullptr );
		delete m_data;
	}
}


AffineTransform GraphicsContext::getCTM( IncludeDeviceScale includeScale ) const
{
	if ( paintingDisabled() )
	{
		return AffineTransform();
	}

	if ( m_impl )
	{
		return m_impl->getCTM( includeScale );
	}

	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	auto ctm = platformContext()->getSceneBuilder()->GetCTM();
	return ConvertFromWebCoreMatrixToTSRenderMatrix( ctm );
}


PlatformContextTSRender *GraphicsContext::platformContext() const
{
	return m_data->platformContext;
}


void GraphicsContext::savePlatformState()
{
	ASSERT( platformContext() );
	platformContext()->save();
}


void GraphicsContext::restorePlatformState()
{
	ASSERT( platformContext() );
	platformContext()->restore();
}


void GraphicsContext::drawRect( const FloatRect &rect, float borderThickness )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->drawRect( rect, borderThickness );
		return;
	}

	ASSERT( !rect.isEmpty() );
	ASSERT( hasPlatformContext() );
	fillRect( rect, fillColor() );

	if ( strokeStyle() != NoStroke )
	{
		strokeRect( rect, borderThickness );
	}
}


void GraphicsContext::drawNativeImage( const NativeImagePtr &image, const FloatSize &imageSize, const FloatRect &destRect, const FloatRect &srcRect, const ImagePaintingOptions &options )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->drawNativeImage( image, imageSize, destRect, srcRect, options );
		return;
	}

	WebCore::drawNativeImage(
		image,
		*this,
		destRect,
		srcRect,
		IntSize( static_cast<int>( imageSize.width() ), static_cast<int>( imageSize.height() ) ),
		options );
}


void GraphicsContext::drawLine( const FloatPoint &point1, const FloatPoint &point2 )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( strokeStyle() == NoStroke )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->drawLine( point1, point2 );
		return;
	}

	auto path = tsRender::Path::Create();
	path->MoveTo( ConvertPointToTSRender( point1 ) );
	path->LineTo( ConvertPointToTSRender( point2 ) );
	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	platformContext()->getSceneBuilder()->StrokePath( *path.get(), strokeThickness(), ConvertColorToTSRender( strokeColor() ) );
}


void GraphicsContext::drawEllipse( const FloatRect &rect )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->drawEllipse( rect );
		return;
	}

	auto path = tsRender::Path::Create();
	auto tsRect = tsRender::FloatRect( rect.x(), rect.y(), rect.x() + rect.width(), rect.y() + rect.height() );
	float rx = rect.width() / 2, ry = rect.height() / 2;
	tsRender::Point pt( rect.x() + rx, rect.y() + ry );
	path->Ellipse( pt, rx, ry );
	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	platformContext()->getSceneBuilder()->FillPath( *path.get(),
		ConvertColorToTSRender( fillColor() ),
		fillRule() == WindRule::NonZero ? tsRender::FillRule::NonZero : tsRender::FillRule::EvenOdd );

	if ( strokeStyle() != NoStroke )
	{
		platformContext()->getSceneBuilder()->StrokePath( *path.get(), strokeThickness(), ConvertColorToTSRender( strokeColor() ) );
	}
}


void GraphicsContext::fillPath( const Path &path )
{
	if ( paintingDisabled() || path.isEmpty() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->fillPath( path );
		return;
	}

	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	tsRender::RefPtr<tsRender::Paint> paint{};
	if ( m_state.fillGradient )
	{
		paint = m_state.fillGradient->createPlatformGradient( m_state.alpha );
	}
	else if ( m_state.fillPattern )
	{
		paint = m_state.fillPattern->createPlatformPattern( AffineTransform() );
	}
	else
	{
		auto color = ConvertColorToTSRender( fillColor() );
		color.a *= m_state.alpha;
		paint = tsRender::SolidPaint::Create( color );
	}

	if ( !paint )
	{
		return;
	}
	
	auto pathFillRule = fillRule() == WindRule::NonZero ? tsRender::FillRule::NonZero : tsRender::FillRule::EvenOdd;
	platformContext()->getSceneBuilder()->FillPath(
		*const_cast<Path &>( path ).ensurePlatformPath(),
		*paint, pathFillRule );
}


void GraphicsContext::strokePath( const Path &path )
{
	if ( paintingDisabled() || path.isEmpty() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->strokePath( path );
		return;
	}

	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	if ( strokeColor().isVisible() && alpha() > 0.f )
	{
		auto color = ConvertColorToTSRender( strokeColor() );
		color.a = static_cast<uint8_t>( alpha() * color.a );
		platformContext()->getSceneBuilder()->StrokePath(
			*const_cast<Path &>( path ).ensurePlatformPath(), strokeThickness(), color,
			ConvertLineJoinToTSRender( platformContext()->lineJoin() ), ConvertLineCapToTSRender( platformContext()->lineCap() ),
			platformContext()->miterLimit(), platformContext()->getDashLineConfig() );
	}
}


void GraphicsContext::fillRect( const FloatRect &rect )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->fillRect( rect );
		return;
	}

	if ( m_state.fillGradient )
	{
		ASSERT( hasPlatformContext() );

		platformContext()->save();
		m_state.fillGradient->fill( *this, rect );
		platformContext()->restore();
		return;
	}
	fillRect( rect, fillColor() );
}


void GraphicsContext::fillRect( const FloatRect &rect, const Color &color )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->fillRect( rect, color );
		return;
	}

	if ( hasVisibleShadow() )
	{
		ShadowBlur contextShadow( m_state );
		contextShadow.drawRectShadow( *this, FloatRoundedRect( rect ) );
	}

	if ( color.isVisible() && alpha() > 0.f )
	{
		auto path = tsRender::Path::Create();
		path->Rect( ConvertFloatRectToTSRender( rect ) );
		ASSERT( platformContext() && platformContext()->getSceneBuilder() );
		auto tsColor = ConvertColorToTSRender( color );
		tsColor.a = static_cast<uint8_t>( tsColor.a * alpha() );
		platformContext()->getSceneBuilder()->FillPath(
			*path.get(), tsColor,
			fillRule() == WindRule::NonZero ? tsRender::FillRule::NonZero : tsRender::FillRule::EvenOdd );
	}
}


void GraphicsContext::clip( const FloatRect &rect )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->clip( rect );
		return;
	}

	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	platformContext()->getSceneBuilder()->Clip( ConvertFloatRectToTSRender( rect ) );
}


void GraphicsContext::clipOut( const FloatRect &rect )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->clipOut( rect );
		return;
	}

	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	platformContext()->getSceneBuilder()->Clip( ConvertFloatRectToTSRender( rect ), true );
}


void GraphicsContext::clipPath( const Path &path, WindRule clipRule )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->clipPath( path, clipRule );
		return;
	}

	if ( path.platformPath() )
	{
		platformContext()->getSceneBuilder()->ClipPath( *path.platformPath(),  clipRule == WindRule::NonZero ? tsRender::FillRule::NonZero : tsRender::FillRule::EvenOdd );
	}
}


void GraphicsContext::clipOut( const Path &path )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->clipOut( path );
		return;
	}

	// We are not actually support path clipping, using path bounding for now
	//auto rect = path.fastBoundingRect();
	//platformContext()->getSceneBuilder()->Clip( ConvertFloatRectToTSRender( rect ), true );
}


void GraphicsContext::clipRoundedRect( const FloatRoundedRect &rect )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		Path path;
		path.addRoundedRect( rect );
		clipPath( path );
		return;
	}
	platformContext()->getSceneBuilder()->ClipRoundedRect( ConvertRoundedRectToTSRender( rect ) );
}


void GraphicsContext::clipOutRoundedRect( const FloatRoundedRect &rect )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( !rect.isRounded() )
	{
		clipOut( rect.rect() );
		return;
	}

	if ( m_impl )
	{
		Path path;
		path.addRoundedRect( rect );
		clipOut( path );
		return;
	}

	platformContext()->getSceneBuilder()->ClipRoundedRect( ConvertRoundedRectToTSRender( rect ), true );
}


void GraphicsContext::clipToImageBuffer( ImageBuffer &buffer, const FloatRect &destRect )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->clipToImageBuffer( buffer, destRect );
		return;
	}

	notImplemented();
}


IntRect GraphicsContext::clipBounds() const
{
	if ( paintingDisabled() )
	{
		return IntRect();
	}

	if ( m_impl )
	{
		return m_impl->clipBounds();
	}

	auto clipBounds = platformContext()->getSceneBuilder()->GetClipBound();
	IntPoint location = { static_cast<int>( clipBounds.Left() ), static_cast<int>( clipBounds.Top() ) };
	IntSize size = { static_cast<int>( clipBounds.Width() ) , static_cast<int>( clipBounds.Height() ) };
	return IntRect( location, size );
}


void GraphicsContext::drawFocusRing( const Path &path, float width, float offset, const Color &color )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->drawFocusRing( path, width, offset, color );
		return;
	}

	Color ringColor = color;
	adjustFocusRingColor( ringColor );
	adjustFocusRingLineWidth( width );
	if ( width <= 0.1 || path.isEmpty() )
	{
		return;
	}

	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	platformContext()->getSceneBuilder()->StrokePath(
		*const_cast<Path &>( path ).ensurePlatformPath(), width * 0.5 / scaleFactor().width(),
		ConvertColorToTSRender( ringColor ),
		ConvertLineJoinToTSRender( platformContext()->lineJoin() ), ConvertLineCapToTSRender( platformContext()->lineCap() ),
		platformContext()->miterLimit() );
}


void GraphicsContext::drawFocusRing( const Vector<FloatRect> &rects, float width, float offset, const Color &color )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->drawFocusRing( rects, width, offset, color );
		return;
	}

	unsigned rectCount = rects.size();
	int radius = ( width - 1 ) / 2;
	if ( rectCount == 0 )
	{
		return;
	}

	FloatRect boundingRect = rects[0];
	for ( unsigned i = 1; i < rectCount; ++i )
		boundingRect.unite( rects[i] );

	Path path;
	path.addRoundedRect( boundingRect, FloatSize( radius, radius ) );
	drawFocusRing( path, width, 0, color );
}


void GraphicsContext::drawLineForText( const FloatRect &rect, bool printing, bool doubleUnderlines, StrokeStyle )
{
	drawLinesForText( rect.location(), rect.height(), DashArray{ 0, rect.width() }, printing, doubleUnderlines );
}


void GraphicsContext::drawLinesForText( const FloatPoint &point, float thickness, const DashArray &widths, bool printing, bool doubleUnderlines, StrokeStyle strokeStyle )
{
	if ( paintingDisabled() )
	{
		return;
	}
	if ( widths.isEmpty() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->drawLinesForText( point, thickness, widths, printing, doubleUnderlines );
		return;
	}

	Color localStrokeColor( strokeColor() );
	FloatRect bounds = computeLineBoundsAndAntialiasingModeForText( FloatRect( point, FloatSize( widths.last(), thickness ) ), printing, localStrokeColor );
	Vector<FloatRect, 4> dashBounds;
	ASSERT( !( widths.size() % 2 ) );
	dashBounds.reserveInitialCapacity( dashBounds.size() / 2 );

	float dashWidth = 0;
	switch ( strokeStyle )
	{
	case DottedStroke:
		dashWidth = bounds.height();
		break;
	case DashedStroke:
		dashWidth = 2 * bounds.height();
		break;
	case SolidStroke:
	default:
		break;
	}

	for ( size_t i = 0; i < widths.size(); i += 2 )
	{
		auto left = widths[i];
		auto width = widths[i + 1] - widths[i];
		if ( !dashWidth )
		{
			dashBounds.append( FloatRect( FloatPoint( bounds.x() + left, bounds.y() ), FloatSize( width, bounds.height() ) ) );
		}
		else
		{
			auto startParticle = static_cast<unsigned>( std::ceil( left / ( 2 * dashWidth ) ) );
			auto endParticle = static_cast<unsigned>( ( left + width ) / ( 2 * dashWidth ) );
			for ( unsigned j = startParticle; j < endParticle; ++j )
			{
				dashBounds.append( FloatRect( FloatPoint( bounds.x() + j * 2 * dashWidth, bounds.y() ), FloatSize( dashWidth, bounds.height() ) ) );
			}
		}
	}

	if ( doubleUnderlines )
	{
		for ( size_t i = 0; i < widths.size(); i += 2 )
		{
			dashBounds.append( FloatRect( FloatPoint( bounds.x() + widths[i], bounds.y() + 2 * bounds.height() ), FloatSize( widths[i + 1] - widths[i], bounds.height() ) ) );
		}
	}

	if ( dashBounds.size() > 0 )
	{
		auto path = tsRender::Path::Create();
		
		for( const auto& rect : dashBounds )
		{
			path->Rect( ConvertFloatRectToTSRender( rect ) );
		}

		platformContext()->getSceneBuilder()->FillPath( *path, ConvertColorToTSRender( strokeColor() ) );
	}
}


void GraphicsContext::drawDotsForDocumentMarker( const FloatRect &rect, DocumentMarkerLineStyle style )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->drawDotsForDocumentMarker( rect, style );
		return;
	}

	notImplemented();
}


FloatRect GraphicsContext::roundToDevicePixels( const FloatRect &rect, RoundingMode roundingMode )
{
	if ( paintingDisabled() )
	{
		return rect;
	}

	if ( m_impl )
	{
		return m_impl->roundToDevicePixels( rect, roundingMode );
	}

	notImplemented();
	return rect;
}


void GraphicsContext::translate( float x, float y )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->translate( x, y );
		return;
	}

	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	platformContext()->getSceneBuilder()->Translate( tsRender::Vec2( x, y ) );
}


void GraphicsContext::concatCTM( const AffineTransform &transform )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->concatCTM( transform );
		return;
	}

	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	platformContext()->getSceneBuilder()->AppendCTM( ConvertAffineTransformToTSRender( transform ) );
}


void GraphicsContext::setCTM( const AffineTransform &transform )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->setCTM( transform );
		return;
	}

	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	platformContext()->getSceneBuilder()->SetCTM( ConvertAffineTransformToTSRender( transform ) );
}


void GraphicsContext::setPlatformShadow( const FloatSize &offset, float, const Color & )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_state.shadowsIgnoreTransforms )
	{
		m_state.shadowOffset = FloatSize( offset.width(), -offset.height() );
	}

	notImplemented();
}


void GraphicsContext::clearPlatformShadow()
{
	notImplemented();
}


void GraphicsContext::beginPlatformTransparencyLayer( float opacity )
{
	if ( paintingDisabled() )
	{
		return;
	}
	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	platformContext()->getSceneBuilder()->BeginTransparentLayer( opacity );
}


void GraphicsContext::endPlatformTransparencyLayer()
{
	if ( paintingDisabled() )
	{
		return;
	}
	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	platformContext()->getSceneBuilder()->EndTransparentLayer();
}


bool GraphicsContext::supportsTransparencyLayers()
{
	return true;
}


void GraphicsContext::clearRect( const FloatRect &rect )
{
	if ( paintingDisabled() )
	{
		return;
	}
	if ( m_impl )
	{
		m_impl->clearRect( rect );
		return;
	}

	platformContext()->getSceneBuilder()->Clear();
}


void GraphicsContext::strokeRect( const FloatRect &rect, float lineWidth )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->strokeRect( rect, lineWidth );
		return;
	}

	if ( lineWidth < 0.001f )
	{
		return;
	}
	if ( rect.isEmpty() || rect.isZero() )
	{
		return;
	}

	if ( strokeColor().isVisible() && alpha() > 0.f )
	{
		Path path;
		path.addRect( rect );
		ASSERT( platformContext() && platformContext()->getSceneBuilder() );
		auto color = ConvertColorToTSRender( strokeColor() );
		color.a = static_cast<uint8_t>( alpha() * color.a );
		platformContext()->getSceneBuilder()->StrokePath(
		*const_cast<Path &>( path ).ensurePlatformPath(), lineWidth, color,
		ConvertLineJoinToTSRender( platformContext()->lineJoin() ), ConvertLineCapToTSRender( platformContext()->lineCap() ),
		platformContext()->miterLimit() );
	}
}


void GraphicsContext::setLineCap( LineCap lineCap )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->setLineCap( lineCap );
		return;
	}

	ASSERT( hasPlatformContext() );

	platformContext()->setLineCap( lineCap );
}


void GraphicsContext::setLineDash( const DashArray &dashes, float dashOffset )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->setLineDash( dashes, dashOffset );
		return;
	}

	ASSERT( hasPlatformContext() );

	platformContext()->setLineDash( dashes, dashOffset );
}


void GraphicsContext::setLineJoin( LineJoin lineJoin )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->setLineJoin( lineJoin );
		return;
	}
	ASSERT( hasPlatformContext() );

	platformContext()->setLineJoin( lineJoin );
}


void GraphicsContext::setMiterLimit( float miter )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->setMiterLimit( miter );
		return;
	}

	ASSERT( hasPlatformContext() );

	platformContext()->setMiterLimit( miter );
}


void GraphicsContext::setPlatformCompositeOperation( CompositeOperator compositeOperator, BlendMode blendMode )
{
	notImplemented();
}


void GraphicsContext::canvasClip( const Path &path, WindRule windRule )
{
	clipPath( path, windRule );
}


void GraphicsContext::rotate( float radians )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->rotate( radians );
		return;
	}

	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	platformContext()->getSceneBuilder()->Rotate( radians );
}


void GraphicsContext::scale( const FloatSize &size )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->scale( size );
		return;
	}

	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	platformContext()->getSceneBuilder()->Scale( size.width(), size.height() );
}


void GraphicsContext::platformFillRoundedRect( const FloatRoundedRect &rect, const Color &color )
{
	if ( paintingDisabled() )
	{
		return;
	}
	if ( hasVisibleShadow() )
	{
		ShadowBlur contextShadow( m_state );
		contextShadow.drawRectShadow( *this, rect );
	}

	auto path = tsRender::Path::Create();
	path->RoundedRect( ConvertRoundedRectToTSRender( rect ) );
	ASSERT( platformContext() && platformContext()->getSceneBuilder() );
	platformContext()->getSceneBuilder()->FillPath(
		*path.get(), ConvertColorToTSRender( color ),
		fillRule() == WindRule::NonZero ? tsRender::FillRule::NonZero : tsRender::FillRule::EvenOdd );
}


void GraphicsContext::fillRectWithRoundedHole( const FloatRect &rect, const FloatRoundedRect &roundedHoleRect, const Color &color )
{
	if ( paintingDisabled() )
	{
		return;
	}
	if ( m_impl )
	{
		m_impl->fillRectWithRoundedHole( rect, roundedHoleRect, color );
		return;
	}

	Path path;
	path.addRect( rect );
	if ( !roundedHoleRect.radii().isZero() )
	{
		path.addRoundedRect( roundedHoleRect );
	}
	else
	{
		path.addRect( roundedHoleRect.rect() );
	}

	WindRule oldFillRule = fillRule();
	Color oldFillColor = fillColor();
	setFillRule( WindRule::EvenOdd );
	setFillColor( color );
	if ( hasVisibleShadow() )
	{
		ShadowBlur contextShadow( m_state );
		contextShadow.drawInsetShadow( *this, rect, roundedHoleRect );
	}
	fillPath( path );
	setFillRule( oldFillRule );
	setFillColor( oldFillColor );
}


void GraphicsContext::drawPattern( Image &image, const FloatRect &destRect, const FloatRect &tileRect, const AffineTransform &patternTransform, const FloatPoint &phase, const FloatSize &spacing, const ImagePaintingOptions &options )
{
	if ( paintingDisabled() )
	{
		return;
	}

	if ( m_impl )
	{
		m_impl->drawPattern( image, destRect, tileRect, patternTransform, phase, spacing, options );
		return;
	}

	if ( !std::isfinite( phase.x() ) || !std::isfinite( phase.y() ) )
	{
		return;
	}

	if ( image.isBitmapImage() )
	{
		BitmapImage &bitmap_image = (BitmapImage &)image;
		NativeImagePtr nativeImage = bitmap_image.nativeImageForCurrentFrame();
		if ( !nativeImage )
		{
			return;
		}

		auto dx = destRect.x(), dy = destRect.y(), dw = destRect.width(), dh = destRect.height();
		auto p2 = tsRender::Path::Create();
		p2->Rect( tsRender::FloatRect::FromSize( dx, dy, dw, dh ) );
		float tileWidth = tileRect.width() * patternTransform.xScale();
		float tileHeight = tileRect.height() * patternTransform.yScale();
		auto srcRect = ConvertFloatRectToTSRender( tileRect );
		auto paint = tsRender::ImagePaint::Create( { phase.x(), phase.y() }, { tileWidth, tileHeight }, nativeImage, srcRect, tsRender::ColorFromRGBA( 255, 255, 255, 255 ) );

		ASSERT( platformContext() && platformContext()->getSceneBuilder() );
		platformContext()->getSceneBuilder()->FillPath( *p2, *paint );
	}
}


void GraphicsContext::setPlatformAlpha( float alpha )
{
	// Don't need to implement
}


void GraphicsContext::setPlatformFillColor( const Color &color )
{
	// Don't need to implement
}


void GraphicsContext::setPlatformStrokeColor( const Color &color )
{
	// Don't need to implement
}


void GraphicsContext::setPlatformStrokeThickness( float strokeThickness )
{
	// Don't need to implement
}


void GraphicsContext::setPlatformShouldAntialias( bool enable )
{
	// Don't need to implement
}


void GraphicsContext::setPlatformImageInterpolationQuality( InterpolationQuality quality )
{
	// Don't need to implement
}


void GraphicsContext::setPlatformStrokeStyle( StrokeStyle strokeStyle )
{
	notImplemented();
}


void GraphicsContext::setURLForRect( const URL &, const FloatRect & )
{
	notImplemented();
}


bool GraphicsContext::isAcceleratedContext() const
{
	return false;
}

} // namespace WebCore

#endif // #if USE( TSRENDER )
