/*
 * Copyright (C) 2020 Activision Publishing, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * 3. Neither the name of Activision Publishing, Inc. nor the names of
 *    its contributors may be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "Path.h"
#include "TSRender.h"
#include "GraphicsContextImpl.h"
#include "NotImplemented.h"
#include "StrokeStyleApplier.h"

#if USE( TSRENDER )
namespace WebCore
{

Path::Path()
	: m_path( nullptr )
{
}


Path::~Path()
{
	m_path = nullptr;
}


Path::Path( const Path &other )
	: m_path( nullptr )
{
	if ( other.isNull() )
	{
		return;
	}
	ensurePlatformPath()->Append( *other.platformPath().get() );
}


Path::Path( Path &&other )
{
	m_path = other.m_path;
	other.m_path = nullptr;
}


Path &Path::operator=( const Path &other )
{
	if ( &other == this )
	{
		return *this;
	}

	if ( other.isNull() )
	{
		m_path = nullptr;
	}
	else
	{
		ensurePlatformPath()->Append( *other.platformPath().get() );
	}
	return *this;
}


PlatformPathPtr Path::ensurePlatformPath()
{
	if ( !m_path )
	{
		m_path = tsRender::Path::Create();
	}
	return m_path;
}


Path &Path::operator=( Path &&other )
{
	if ( this == &other )
	{
		return *this;
	}
	m_path = other.m_path;
	other.m_path = nullptr;
	return *this;
}

void Path::clear()
{
	if ( isNull() )
	{
		return;
	}
	m_path = tsRender::Path::Create();
}


bool Path::isEmpty() const
{
	return isNull() || platformPath()->IsEmpty();
}


bool Path::hasCurrentPoint() const
{
	if ( isNull() )
	{
		return false;
	}
	return platformPath()->HasCurrentPoint();
}


FloatPoint Path::currentPoint() const
{
	if ( isNull() )
	{
		return FloatPoint();
	}
	tsRender::Point p = platformPath()->GetCurrentPoint();
	return FloatPoint( p.x, p.y );
}


void Path::translate( const FloatSize &p )
{
	transform( AffineTransform( 1, 0, 0, 1, p.width(), p.height() ) );
}


void Path::moveTo( const FloatPoint &p )
{
	ensurePlatformPath()->MoveTo( { p.x(), p.y() } );
}


void Path::addLineTo( const FloatPoint &p )
{
	ensurePlatformPath()->LineTo( { p.x(), p.y() } );
}


void Path::addRect( const FloatRect &rect )
{
	auto path = ensurePlatformPath();
	path->MoveTo( { rect.x(), rect.y() } );
	path->LineTo( { rect.x() + rect.width(), rect.y() } );
	path->LineTo( { rect.x() + rect.width(), rect.y() + rect.height() } );
	path->LineTo( { rect.x(), rect.y() + rect.height() } );
	path->Close();
}


void Path::addQuadCurveTo( const FloatPoint &controlPoint, const FloatPoint &point )
{
	ensurePlatformPath()->QuadCurveTo( { controlPoint.x(), controlPoint.y() },
		{ point.x(), point.y() } );
}


void Path::addBezierCurveTo( const FloatPoint &controlPoint1, const FloatPoint &controlPoint2, const FloatPoint &controlPoint3 )
{
	ensurePlatformPath()->BezierCurveTo( { controlPoint1.x(), controlPoint1.y() },
		{ controlPoint2.x(), controlPoint2.y() },
		{ controlPoint3.x(), controlPoint3.y() } );
}


void Path::addArc( const FloatPoint &p, float r, float startAngle, float endAngle, bool anticlockwise )
{
	if ( !std::isfinite( r ) || !std::isfinite( startAngle ) || !std::isfinite( endAngle ) )
	{
		return;
	}

	auto path = ensurePlatformPath();
	float sweep = endAngle - startAngle;
	const float twoPI = 2 * piFloat;
	if ( ( sweep <= -twoPI || sweep >= twoPI )
		&& ( ( anticlockwise && ( endAngle < startAngle ) ) || ( !anticlockwise && ( startAngle < endAngle ) ) ) )
	{
		if ( anticlockwise )
		{
			path->ArcTo( { p.x(), p.y() }, r, startAngle, startAngle - twoPI, true );
		}
		else
		{
			path->ArcTo( { p.x(), p.y() }, r, startAngle, startAngle + twoPI, false );
		}
		path->Close();
		path->ArcTo( { p.x(), p.y() }, r, endAngle, endAngle, false );
	}
	else
	{
		if ( anticlockwise )
		{
			path->ArcTo( { p.x(), p.y() }, r, startAngle, endAngle, true );
		}
		else
		{
			path->ArcTo( { p.x(), p.y() }, r, startAngle, endAngle, false );
		}
	}
}


static inline float areaOfTriangleFormedByPoints( const FloatPoint &p1, const FloatPoint &p2, const FloatPoint &p3 )
{
	return p1.x() * ( p2.y() - p3.y() ) + p2.x() * ( p3.y() - p1.y() ) + p3.x() * ( p1.y() - p2.y() );
}


void Path::addArcTo( const FloatPoint &p1, const FloatPoint &p2, float radius )
{
	auto path = ensurePlatformPath();
	tsRender::Point cur_p = path->GetCurrentPoint();
	FloatPoint p0( cur_p.x, cur_p.y );
	if ( ( p1.x() == p0.x() && p1.y() == p0.y() ) || ( p1.x() == p2.x() && p1.y() == p2.y() ) || !radius
		|| !areaOfTriangleFormedByPoints( p0, p1, p2 ) )
	{
		path->LineTo( { p1.x(), p1.y() } );
		return;
	}

	FloatPoint p1p0( ( p0.x() - p1.x() ), ( p0.y() - p1.y() ) );
	FloatPoint p1p2( ( p2.x() - p1.x() ), ( p2.y() - p1.y() ) );
	float p1p0_length = sqrtf( p1p0.x() * p1p0.x() + p1p0.y() * p1p0.y() );
	float p1p2_length = sqrtf( p1p2.x() * p1p2.x() + p1p2.y() * p1p2.y() );

	double cos_phi = ( p1p0.x() * p1p2.x() + p1p0.y() * p1p2.y() ) / ( p1p0_length * p1p2_length );
	if ( cos_phi == -1 )
	{
		path->LineTo( { p1.x(), p1.y() } );
		return;
	}
	if ( cos_phi == 1 )
	{
		unsigned int max_length = 65535;
		double factor_max = max_length / p1p0_length;
		FloatPoint ep( ( p0.x() + factor_max * p1p0.x() ), ( p0.y() + factor_max * p1p0.y() ) );
		path->LineTo( { ep.x(), ep.y() } );
		return;
	}

	float tangent = radius / tan( acos( cos_phi ) / 2 );
	float factor_p1p0 = tangent / p1p0_length;
	FloatPoint t_p1p0( ( p1.x() + factor_p1p0 * p1p0.x() ), ( p1.y() + factor_p1p0 * p1p0.y() ) );

	FloatPoint orth_p1p0( p1p0.y(), -p1p0.x() );
	float orth_p1p0_length = sqrt( orth_p1p0.x() * orth_p1p0.x() + orth_p1p0.y() * orth_p1p0.y() );
	float factor_ra = radius / orth_p1p0_length;

	double cos_alpha = ( orth_p1p0.x() * p1p2.x() + orth_p1p0.y() * p1p2.y() ) / ( orth_p1p0_length * p1p2_length );
	if ( cos_alpha < 0.f )
	{
		orth_p1p0 = FloatPoint( -orth_p1p0.x(), -orth_p1p0.y() );
	}

	FloatPoint p( ( t_p1p0.x() + factor_ra * orth_p1p0.x() ), ( t_p1p0.y() + factor_ra * orth_p1p0.y() ) );

	orth_p1p0 = FloatPoint( -orth_p1p0.x(), -orth_p1p0.y() );
	float sa = acos( orth_p1p0.x() / orth_p1p0_length );
	if ( orth_p1p0.y() < 0.f )
	{
		sa = 2 * piDouble - sa;
	}

	bool anticlockwise = false;

	float factor_p1p2 = tangent / p1p2_length;
	FloatPoint t_p1p2( ( p1.x() + factor_p1p2 * p1p2.x() ), ( p1.y() + factor_p1p2 * p1p2.y() ) );
	FloatPoint orth_p1p2( ( t_p1p2.x() - p.x() ), ( t_p1p2.y() - p.y() ) );
	float orth_p1p2_length = sqrtf( orth_p1p2.x() * orth_p1p2.x() + orth_p1p2.y() * orth_p1p2.y() );
	float ea = acos( orth_p1p2.x() / orth_p1p2_length );
	if ( orth_p1p2.y() < 0 )
	{
		ea = 2 * piDouble - ea;
	}
	if ( ( sa > ea ) && ( ( sa - ea ) < piDouble ) )
	{
		anticlockwise = true;
	}
	if ( ( sa < ea ) && ( ( ea - sa ) > piDouble ) )
	{
		anticlockwise = true;
	}

	path->LineTo( { t_p1p0.x(), t_p1p0.y() } );
	addArc( p, radius, sa, ea, anticlockwise );
}


void Path::addEllipse( FloatPoint point, float radiusX, float radiusY, float rotation, float startAngle, float endAngle, bool anticlockwise )
{
	auto path = ensurePlatformPath();
	path->Ellipse( ConvertPointToTSRender( point ), radiusX, radiusY );
}


void Path::addEllipse( const FloatRect &rect )
{
	auto path = ensurePlatformPath();
	tsRender::Point center = tsRender::Point( rect.x() + rect.width() / 2, rect.y() + rect.height() / 2 );
	float yRadius = .5 * rect.height();
	float xRadius = .5 * rect.width();
	path->Ellipse( center, xRadius, yRadius );
}


void Path::addPath( const Path &path, const AffineTransform &transform )
{
	if ( path.platformPath() )
	{
		ensurePlatformPath()->Append( *path.platformPath().get() );
	}
}


void Path::closeSubpath()
{
	ensurePlatformPath()->Close();
}


FloatRect Path::boundingRect() const
{
	if ( isNull() )
	{
		return FloatRect();
	}

	auto aabb = platformPath()->GetBoundingBox();
	return FloatRect( aabb.Left(), aabb.Top(), aabb.Right() - aabb.Left(), aabb.Bottom() - aabb.Top() );
}


FloatRect Path::strokeBoundingRect( StrokeStyleApplier *applier ) const
{
	if ( isNull() )
	{
		return FloatRect();
	}

	if ( !applier )
	{
		return FloatRect();
	}
	GraphicsContext context{};
	applier->strokeStyle( &context );
	float strokeWidth = context.strokeThickness();
	auto strokeBound = boundingRect();
	strokeBound.inflate( strokeWidth );
	return strokeBound;
}


bool Path::contains( const FloatPoint &point, WindRule rule ) const
{
	if ( isNull() || !std::isfinite( point.x() ) || !std::isfinite( point.y() ) )
	{
		return false;
	}
	auto path = platformPath();
	return path->IsPointFilled( { point.x(), point.y() } );
}


bool Path::strokeContains( StrokeStyleApplier &applier, const FloatPoint &point ) const
{
	if ( isNull() )
	{
		return false;
	}
	return contains( point );
}


void Path::apply( const PathApplierFunction &function ) const
{
	notImplemented();
}


void Path::transform( const AffineTransform &transform )
{
	tsRender::AffineTransform trans( transform.a(), transform.b(), transform.c(), transform.d(), transform.e(), transform.f() );
	ensurePlatformPath()->Transform( trans );
}


} // namespace WebCore
#endif // #if USE( TSRENDER )
