/*
 * Copyright (C) 2006-2017 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "WebFrameLoaderClient.h"

#include "WebChromeClient.h"
#include "WebFrame.h"
#include "WebView.h"
#include "WebFrameNetworkingContext.h"
#include "ClientCallback.h"
#include <JavaScriptCore/APICast.h>
#include <WebCore/BackForwardController.h>
#include <WebCore/CachedFrame.h>
#include <WebCore/DNS.h>
#include <WebCore/DocumentLoader.h>
#include <WebCore/FormState.h>
#include <WebCore/Frame.h>
#include <WebCore/FrameLoader.h>
#include <WebCore/FrameTree.h>
#include <WebCore/FrameView.h>
#include <WebCore/HTMLAppletElement.h>
#include <WebCore/HTMLFrameElement.h>
#include <WebCore/HTMLFrameOwnerElement.h>
#include <WebCore/HTMLNames.h>
#include <WebCore/HTMLParserIdioms.h>
#include <WebCore/HTMLPlugInElement.h>
#include <WebCore/HistoryItem.h>
#include <WebCore/LocalizedStrings.h>
#include <WebCore/MIMETypeRegistry.h>
#include <WebCore/NotImplemented.h>
#include <WebCore/Page.h>
#include <WebCore/PolicyChecker.h>
#include <WebCore/RenderWidget.h>
#include <WebCore/ResourceHandle.h>
#include <WebCore/ScriptController.h>
#include <WebCore/Settings.h>
#include <WebCore/SubresourceLoader.h>
#include <webcore/FrameLoaderTypes.h>
#include "Universal/TelescopeLog.h"
#include "ClientWrappers/LogReplacement/LogReplacement.h"
#include "WebCoreSupport/WebFrameLoadDelegate.h"
#include <WebCore/MemoryRelease.h>

using namespace WebCore;
using namespace HTMLNames;
using namespace Telescope;

// When the web view reloaded we would want to aggressively release memory, 
// this was done in a timer, which will be called when next time
// telescope tick, the reason don't do it immediately is because a lot of things was cleaned
// in timer which is in next time telescope tick, if we clean up immediately, a lot of stuff
// were not released.
static void AggressivelyReleaseMemory()
{
    releaseMemory( Critical::Yes, Synchronous::Yes );
}

WebFrameLoaderClient::WebFrameLoaderClient(WebFrame* webFrame, ClientCallbackInterface* systemCallback)
    : m_webFrame(webFrame), m_sysCB(systemCallback)
    ,m_timer( AggressivelyReleaseMemory )
{
}

WebFrameLoaderClient::~WebFrameLoaderClient()
{
}

void WebFrameLoaderClient::frameLoaderDestroyed()
{
    delete m_webFrame;
    m_webFrame = nullptr;

    delete this;
}

Optional<WebCore::PageIdentifier> WebFrameLoaderClient::pageID() const
{
    return WTF::nullopt;
}

Optional<WebCore::FrameIdentifier> WebFrameLoaderClient::frameID() const
{
    return WTF::nullopt;
}

bool WebFrameLoaderClient::hasWebView() const
{
    return m_webFrame->webView();
}

void WebFrameLoaderClient::makeRepresentation(DocumentLoader*)
{
    notImplemented();
}

void WebFrameLoaderClient::forceLayoutForNonHTML()
{
    notImplemented();
}

void WebFrameLoaderClient::setCopiesOnScroll()
{
    notImplemented();
}

void WebFrameLoaderClient::detachedFromParent2()
{
    notImplemented();
}

void WebFrameLoaderClient::detachedFromParent3()
{
    notImplemented();
}

void WebFrameLoaderClient::convertMainResourceLoadToDownload(DocumentLoader* documentLoader, const ResourceRequest& request, const ResourceResponse& response)
{
    // No, we don't support download for telescope
    notImplemented();
}

bool WebFrameLoaderClient::dispatchDidLoadResourceFromMemoryCache(DocumentLoader*, const ResourceRequest&, const ResourceResponse&, int /*length*/)
{
    notImplemented();
    return false;
}

void WebFrameLoaderClient::assignIdentifierToInitialRequest(unsigned long identifier, DocumentLoader* loader, const ResourceRequest& request)
{
    notImplemented();
}

bool WebFrameLoaderClient::shouldUseCredentialStorage(DocumentLoader* loader, unsigned long identifier)
{
    return false;
}

void WebFrameLoaderClient::dispatchDidReceiveAuthenticationChallenge(DocumentLoader* loader, unsigned long identifier, const AuthenticationChallenge& challenge)
{
    notImplemented();
}

void WebFrameLoaderClient::dispatchWillSendRequest(DocumentLoader* loader, unsigned long identifier, ResourceRequest& request, const ResourceResponse& redirectResponse)
{
    // looks like it give a us a chance to manipulate the redirect request 
    // but didn't find any requirement for us to do so
    notImplemented();
}

void WebFrameLoaderClient::dispatchDidReceiveResponse(DocumentLoader* loader, unsigned long identifier, const ResourceResponse& response)
{
    static const char* pSourceText[] =
    {
        "Unknown",
        "Network",
        "DiskCache",
        "DiskCacheAfterValidation",
        "MemoryCache",
        "MemoryCacheAfterValidation",
        "ServiceWorker",
        "ApplicationCache",
        "DOMCache",
        "InspectorOverride"
    };
    static_assert(WTF_ARRAY_LENGTH(pSourceText) == static_cast<size_t>(ResourceResponseBase::Source::InspectorOverride) + 1);

    if (GetLogWrapper()->GetLevel() == LogLevel::Debug)
    {
        LogDebug(LogChannel::GENERAL, "++++++++++++ResourceResponse++++++++++++");

        LogDebug(LogChannel::GENERAL,
            "\n"
            "URL:%s\n"
            "httpStatusCode:%d\n"
            "Source:%s\n"
            , response.url().string().ascii().data()
            , response.httpStatusCode()
            , pSourceText[static_cast<uint8_t>(response.source())] );
      
        constexpr size_t BUFFER_SIZE = 1024;
        char buffer[BUFFER_SIZE] = "";
        size_t bufferUsed = 0;       

        auto iterateHeader = [=, &bufferUsed, &buffer](const char* key, const char* val)
        {
            bool pendingProcess = true;
            do
            {
                size_t requiredLen = strlen(key) + strlen(val) + 3; // 1 for ':', 1 for '\n', 1 for '\0'
                size_t availableLen = std::min(requiredLen, BUFFER_SIZE - bufferUsed);            
                if (availableLen < requiredLen)
                {
                    if (bufferUsed == 0)
                    {
                        snprintf(buffer, BUFFER_SIZE, "\n%s:%s", key, val);
                        pendingProcess = false;
                    }
                    LogDebug(LogChannel::GENERAL, buffer);
                    bufferUsed = 0;
                }
                else
                {
                    bufferUsed += snprintf(buffer + bufferUsed, availableLen, "\n%s:%s", key, val);
                    pendingProcess = false;
                }
            }
            while(pendingProcess);
        };

        auto commonHeaders = response.httpHeaderFields().commonHeaders();
        auto uncommonHeaders = response.httpHeaderFields().uncommonHeaders();
        for (auto it = commonHeaders.begin(); it != commonHeaders.end(); ++it)
        {
            iterateHeader(httpHeaderNameString(it->key).toStringWithoutCopying().ascii().data(), it->value.ascii().data());
        }
        for (auto it = uncommonHeaders.begin(); it != uncommonHeaders.end(); ++it)
        {
            iterateHeader(it->key.ascii().data(), it->value.ascii().data());
        }
        if (bufferUsed > 0)
        {
            LogDebug(LogChannel::GENERAL, buffer);
        }

        LogDebug(LogChannel::GENERAL, "------------ResourceResponse------------");
    }
}

void WebFrameLoaderClient::dispatchDidReceiveContentLength(DocumentLoader* loader, unsigned long identifier, int length)
{
    notImplemented();
}

void WebFrameLoaderClient::dispatchDidFinishLoading(DocumentLoader* loader, unsigned long identifier)
{
    notImplemented();
}

void WebFrameLoaderClient::dispatchDidFailLoading(DocumentLoader* loader, unsigned long identifier, const ResourceError& error)
{
    notImplemented();
}

void WebFrameLoaderClient::dispatchDidDispatchOnloadEvents()
{
	WebView *webView = m_webFrame->webView();
	WebFrameLoadDelegate *frameLoadDelegate = webView->frameLoadDelegate();
	if ( frameLoadDelegate )
	{
		frameLoadDelegate->didHandleOnloadEventsForFrame( webView, m_webFrame );
	}
}

void WebFrameLoaderClient::dispatchDidReceiveServerRedirectForProvisionalLoad()
{
	WebView *webView = m_webFrame->webView();
	WebFrameLoadDelegate *frameLoadDelegate = webView->frameLoadDelegate();
	if ( frameLoadDelegate )
	{
		frameLoadDelegate->didReceiveServerRedirectForProvisionalLoadForFrame( webView, m_webFrame );
	}
}

void WebFrameLoaderClient::dispatchDidCancelClientRedirect()
{
	WebView *webView = m_webFrame->webView();
	WebFrameLoadDelegate *frameLoadDelegate = webView->frameLoadDelegate();
	if ( frameLoadDelegate )
	{
		frameLoadDelegate->didCancelClientRedirectForFrame( webView, m_webFrame );
	}
}

void WebFrameLoaderClient::dispatchWillPerformClientRedirect(const URL& url, double delay, WallTime fireDate, WebCore::LockBackForwardList)
{
	WebView *webView = m_webFrame->webView();
	WebFrameLoadDelegate *frameLoadDelegate = webView->frameLoadDelegate();
	if ( frameLoadDelegate )
	{
		frameLoadDelegate->willPerformClientRedirectToURL( webView, url.string(), delay, fireDate.secondsSinceEpoch().seconds(), m_webFrame );
	}
}

void WebFrameLoaderClient::dispatchDidChangeLocationWithinPage()
{
	WebView *webView = m_webFrame->webView();
	WebFrameLoadDelegate *frameLoadDelegate = webView->frameLoadDelegate();
	if ( frameLoadDelegate )
	{
		frameLoadDelegate->didChangeLocationWithinPageForFrame( webView, m_webFrame );
	}
}

void WebFrameLoaderClient::dispatchDidPushStateWithinPage()
{
    notImplemented();
}

void WebFrameLoaderClient::dispatchDidReplaceStateWithinPage()
{
    notImplemented();
}

void WebFrameLoaderClient::dispatchDidPopStateWithinPage()
{
    notImplemented();
}

void WebFrameLoaderClient::dispatchWillClose()
{
    notImplemented();
}

void WebFrameLoaderClient::dispatchDidStartProvisionalLoad()
{
	WebView *webView = m_webFrame->webView();
	WebFrameLoadDelegate *frameLoadDelegate = webView->frameLoadDelegate();
	if ( frameLoadDelegate )
	{
		frameLoadDelegate->didStartProvisionalLoadForFrame( webView, m_webFrame );
	}
}

void WebFrameLoaderClient::dispatchDidReceiveTitle(const StringWithDirection& title)
{
	WebView *webView = m_webFrame->webView();
	WebFrameLoadDelegate *frameLoadDelegate = webView->frameLoadDelegate();
	if ( frameLoadDelegate )
	{
		frameLoadDelegate->didReceiveTitle( webView, title.string, m_webFrame );
	}
}

void WebFrameLoaderClient::dispatchDidCommitLoad(Optional<HasInsecureContent>)
{
	WebView *webView = m_webFrame->webView();
	WebFrameLoadDelegate *frameLoadDelegate = webView->frameLoadDelegate();
	if ( frameLoadDelegate )
	{
		frameLoadDelegate->didCommitLoadForFrame( webView, m_webFrame );
	}
}

void WebFrameLoaderClient::dispatchDidFailProvisionalLoad(const ResourceError& error, WillContinueLoading)
{
	WebView *webView = m_webFrame->webView();
	WebFrameLoadDelegate *frameLoadDelegate = webView->frameLoadDelegate();
	if ( frameLoadDelegate )
	{
		frameLoadDelegate->didFailProvisionalLoad( webView, m_webFrame );
	}
}

void WebFrameLoaderClient::dispatchDidFailLoad(const ResourceError& error)
{
	WebView *webView = m_webFrame->webView();
	WebFrameLoadDelegate *frameLoadDelegate = webView->frameLoadDelegate();
	if ( frameLoadDelegate )
	{
		frameLoadDelegate->didFailLoad( webView, m_webFrame );
	}
}

void WebFrameLoaderClient::dispatchDidFinishDocumentLoad()
{
    if ( m_sysCB )
    {
        m_sysCB->FinishDocumentLoadHandler();
    }
}

void WebFrameLoaderClient::dispatchDidFinishLoad()
{
	WebView *webView = m_webFrame->webView();
	WebFrameLoadDelegate *frameLoadDelegate = webView->frameLoadDelegate();
	if ( frameLoadDelegate )
	{
		frameLoadDelegate->didFinishLoadForFrame( webView, m_webFrame );
	}
}

void WebFrameLoaderClient::dispatchDidReachLayoutMilestone(OptionSet<WebCore::LayoutMilestone> milestones)
{
    notImplemented();
}

Frame* WebFrameLoaderClient::dispatchCreatePage(const NavigationAction& navigationAction)
{
    // we are not going to create a new page
    notImplemented();
    return nullptr;
}

void WebFrameLoaderClient::dispatchShow()
{
    notImplemented();
}

void WebFrameLoaderClient::dispatchDecidePolicyForResponse(const ResourceResponse& response, const ResourceRequest& request, WebCore::PolicyCheckIdentifier identifier, const String&, FramePolicyFunction&& function)
{
    if (!m_webFrame)
        return;

    function(PolicyAction::Use, identifier);
}

void WebFrameLoaderClient::dispatchDecidePolicyForNewWindowAction(const NavigationAction& action, const ResourceRequest& request, FormState* formState, const String& frameName, WebCore::PolicyCheckIdentifier identifier, FramePolicyFunction&& function)
{
    function(PolicyAction::Use, identifier);
}

void WebFrameLoaderClient::dispatchDecidePolicyForNavigationAction(const NavigationAction& action, const ResourceRequest& request, const ResourceResponse&, FormState* formState, WebCore::PolicyDecisionMode, WebCore::PolicyCheckIdentifier identifier, FramePolicyFunction&& function)
{
    function(PolicyAction::Use, identifier);
}

void WebFrameLoaderClient::dispatchUnableToImplementPolicy(const ResourceError& error)
{
    notImplemented();
}

void WebFrameLoaderClient::dispatchWillSendSubmitEvent(Ref<WebCore::FormState>&&)
{
}

void WebFrameLoaderClient::dispatchWillSubmitForm(FormState& formState, CompletionHandler<void()>&& completionHandler)
{
    // FIXME: Add a sane default implementation
    completionHandler();
}

void WebFrameLoaderClient::setMainDocumentError(DocumentLoader*, const ResourceError& error)
{
}

void WebFrameLoaderClient::startDownload(const ResourceRequest& request, const String& /* suggestedName */)
{
}

void WebFrameLoaderClient::willChangeTitle(DocumentLoader*)
{
    notImplemented();
}

void WebFrameLoaderClient::didChangeTitle(DocumentLoader*)
{
    notImplemented();
}

void WebFrameLoaderClient::committedLoad(DocumentLoader* loader, const char* data, int length)
{
    loader->commitData(data, length);

    // If the document is a stand-alone media document, now is the right time to cancel the WebKit load.
    // FIXME: This code should be shared across all ports. <http://webkit.org/b/48762>.
    Frame* coreFrame = core(m_webFrame);
    if (coreFrame->document()->isMediaDocument())
        loader->cancelMainResourceLoad(pluginWillHandleLoadError(loader->response()));
}

void WebFrameLoaderClient::finishedLoading(DocumentLoader*)
{
}

void WebFrameLoaderClient::updateGlobalHistory()
{
}

void WebFrameLoaderClient::updateGlobalHistoryRedirectLinks()
{
}

bool WebFrameLoaderClient::shouldGoToHistoryItem(HistoryItem&) const
{
    return true;
}

void WebFrameLoaderClient::didDisplayInsecureContent()
{
	WebView *webView = m_webFrame->webView();
	WebFrameLoadDelegate *frameLoadDelegate = webView->frameLoadDelegate();
	if ( frameLoadDelegate )
	{
		frameLoadDelegate->didDisplayInsecureContent( webView );
	}
}

void WebFrameLoaderClient::didRunInsecureContent(SecurityOrigin& origin, const URL& insecureURL)
{
	WebView *webView = m_webFrame->webView();
	WebFrameLoadDelegate *frameLoadDelegate = webView->frameLoadDelegate();
	if ( frameLoadDelegate )
	{
		frameLoadDelegate->didRunInsecureContent( webView );
	}
}

void WebFrameLoaderClient::didDetectXSS(const URL&, bool)
{
    // FIXME: propogate call into the private delegate.
}

ResourceError WebFrameLoaderClient::cancelledError(const ResourceRequest& request)
{
    // FIXME: Need ChickenCat to include CFNetwork/CFURLError.h to get these values
    // Alternatively, we could create our own error domain/codes.
    return ResourceError(String(WebURLErrorDomain), -999, request.url(), String("Cancelled"));
}

ResourceError WebFrameLoaderClient::blockedError(const ResourceRequest& request)
{
    return ResourceError(String(WebKitErrorDomain), WebKitErrorCannotUseRestrictedPort, request.url(), WEB_UI_STRING("Not allowed to use restricted network port", "WebKitErrorCannotUseRestrictedPort description"));
}

ResourceError WebFrameLoaderClient::blockedByContentBlockerError(const ResourceRequest& request)
{
    RELEASE_ASSERT_NOT_REACHED(); // Content Blockers are not enabled for WK1.
}

ResourceError WebFrameLoaderClient::cannotShowURLError(const ResourceRequest& request)
{
    return ResourceError(String(WebKitErrorDomain), WebKitErrorCannotShowURL, request.url(), WEB_UI_STRING("The URL can\xE2\x80\x99t be shown", "WebKitErrorCannotShowURL description"));
}

ResourceError WebFrameLoaderClient::interruptedForPolicyChangeError(const ResourceRequest& request)
{
    return ResourceError(String(WebKitErrorDomain), WebKitErrorFrameLoadInterruptedByPolicyChange, request.url(), WEB_UI_STRING("Frame load interrupted", "WebKitErrorFrameLoadInterruptedByPolicyChange description"));
}

ResourceError WebFrameLoaderClient::cannotShowMIMETypeError(const ResourceResponse& response)
{
    return ResourceError(String(), WebKitErrorCannotShowMIMEType, response.url(), WEB_UI_STRING("Content with specified MIME type can\xE2\x80\x99t be shown", "WebKitErrorCannotShowMIMEType description"));
}

ResourceError WebFrameLoaderClient::fileDoesNotExistError(const ResourceResponse& response)
{
    return ResourceError(String(WebURLErrorDomain), -1100, response.url(), String("File does not exist."));
}

ResourceError WebFrameLoaderClient::pluginWillHandleLoadError(const ResourceResponse& response)
{
    return ResourceError(String(WebKitErrorDomain), WebKitErrorPlugInWillHandleLoad, response.url(), WEB_UI_STRING("Plug-in handled load", "WebKitErrorPlugInWillHandleLoad description"));
}

bool WebFrameLoaderClient::shouldFallBack(const ResourceError& error)
{
    if (error.errorCode() == WebURLErrorCancelled && error.domain() == String(WebURLErrorDomain))
        return false;

    if (error.errorCode() == WebKitErrorPlugInWillHandleLoad && error.domain() == String(WebKitErrorDomain))
        return false;

    return true;
}

bool WebFrameLoaderClient::canHandleRequest(const ResourceRequest& request) const
{
    return WebView::canHandleRequest(&request);
}

static bool isSupportedMIMEType(const String& mimeType)
{
	return MIMETypeRegistry::isSupportedImageMIMEType(mimeType)
		|| MIMETypeRegistry::isSupportedNonImageMIMEType(mimeType)
		|| MIMETypeRegistry::isSupportedMediaMIMEType(mimeType);
}
bool WebFrameLoaderClient::canShowMIMEType(const String& mimeType) const
{
    return isSupportedMIMEType(mimeType);
}

bool WebFrameLoaderClient::canShowMIMETypeAsHTML(const String& mimeType) const
{
    return true;
}

bool WebFrameLoaderClient::representationExistsForURLScheme(const String& /*URLScheme*/) const
{
    notImplemented();
    return false;
}

String WebFrameLoaderClient::generatedMIMETypeForURLScheme(const String& /*URLScheme*/) const
{
    notImplemented();
    ASSERT_NOT_REACHED();
    return String();
}

void WebFrameLoaderClient::frameLoadCompleted()
{
}

void WebFrameLoaderClient::saveViewStateToItem(HistoryItem&)
{
}

void WebFrameLoaderClient::restoreViewState()
{
}

void WebFrameLoaderClient::provisionalLoadStarted()
{
    notImplemented();
}

void WebFrameLoaderClient::didFinishLoad()
{
    notImplemented();
}

void WebFrameLoaderClient::prepareForDataSourceReplacement()
{
    notImplemented();
}

Ref<DocumentLoader> WebFrameLoaderClient::createDocumentLoader(const ResourceRequest& request, const SubstituteData& substituteData)
{
    Ref<DocumentLoader> loader = DocumentLoader::create(request, substituteData);

    return loader;
}

void WebFrameLoaderClient::setTitle(const StringWithDirection& title, const URL& url)
{
}

void WebFrameLoaderClient::savePlatformDataToCachedFrame(CachedFrame* cachedFrame)
{
    notImplemented();
}

void WebFrameLoaderClient::transitionToCommittedFromCachedFrame(CachedFrame*)
{
}

void WebFrameLoaderClient::transitionToCommittedForNewPage()
{
    WebView* view = m_webFrame->webView();

    Optional<Color> backgroundColor;
    if (view->transparent())
        backgroundColor = Color(Color::transparent);
    
    FloatRect logicalFrame = *view->windowRect();;
    core(m_webFrame)->createView(enclosingIntRect(logicalFrame).size(), backgroundColor, /* fixedLayoutSize */ { }, /* fixedVisibleContentRect */ { });
}

void WebFrameLoaderClient::didRestoreFromBackForwardCache()
{
}

void WebFrameLoaderClient::dispatchDidBecomeFrameset(bool)
{
}

static const WTF::String defaultUserAgent()
{
#if TELESCOPE_USING(TELESCOPE_PC_PROGRAM)
    static const char* pPrefixUserAgent = "Telescope.PC.";
#elif TELESCOPE_USING(TELESCOPE_XB3_PROGRAM)
    static const char* pPrefixUserAgent = "Telescope.XB3.";
#elif TELESCOPE_USING(TELESCOPE_XB4_PROGRAM)
    static const char* pPrefixUserAgent = "Telescope.XB4.";
#elif TELESCOPE_USING(TELESCOPE_PS4_PROGRAM)
    static const char* pPrefixUserAgent = "Telescope.PS4.";
#elif TELESCOPE_USING(TELESCOPE_PS5_PROGRAM)
    static const char* pPrefixUserAgent = "Telescope.PS5.";
#else
#error "Unknown platform"
#endif
    static const NeverDestroyed<String> agent = makeString(pPrefixUserAgent, GetChangelistNumberInterface());
    return agent;
}

String WebFrameLoaderClient::userAgent(const URL& url)
{
    return defaultUserAgent();
}

bool WebFrameLoaderClient::canCachePage() const
{
    return true;
}

RefPtr<Frame> WebFrameLoaderClient::createFrame(const URL& url, const String& name, HTMLFrameOwnerElement& ownerElement,
    const String& referrer)
{
    Frame* coreFrame = core(m_webFrame);
    ASSERT(coreFrame);

    WebFrame* webFrame = WebFrame::createInstance();

    RefPtr<Frame> childFrame = webFrame->createSubframeWithOwnerElement(m_webFrame->webView(), coreFrame->page(), &ownerElement);

    childFrame->tree().setName(name);
    coreFrame->tree().appendChild(*childFrame);
    childFrame->init();

    coreFrame->loader().loadURLIntoChildFrame(url, referrer, childFrame.get());

    // The frame's onload handler may have removed it from the document.
    if (!childFrame->tree().parent())
        return nullptr;

    return childFrame;
}

ObjectContentType WebFrameLoaderClient::objectContentType(const URL& url, const String& mimeTypeIn)
{
    String mimeType = mimeTypeIn;

    if (mimeType.isEmpty())
        return ObjectContentType::Frame; // Go ahead and hope that we can display the content.


    if (MIMETypeRegistry::isSupportedImageMIMEType(mimeType))
        return WebCore::ObjectContentType::Image;


    if (MIMETypeRegistry::isSupportedNonImageMIMEType(mimeType))
        return WebCore::ObjectContentType::Frame;

    return WebCore::ObjectContentType::None;
}

void WebFrameLoaderClient::dispatchDidFailToStartPlugin(const PluginView& pluginView) const
{
    ASSERT(0);
}

RefPtr<Widget> WebFrameLoaderClient::createPlugin(const IntSize& pluginSize, HTMLPlugInElement& element, const URL& url, const Vector<String>& paramNames, const Vector<String>& paramValues, const String& mimeType, bool loadManually)
{
    return nullptr;
}

void WebFrameLoaderClient::redirectDataToPlugin(Widget& pluginWidget)
{
    // Ideally, this function shouldn't be necessary, see <rdar://problem/4852889>
}

RefPtr<Widget> WebFrameLoaderClient::createJavaAppletWidget(const IntSize& pluginSize, HTMLAppletElement& element, const URL& /*baseURL*/, const Vector<String>& paramNames, const Vector<String>& paramValues)
{
    return nullptr;
}

String WebFrameLoaderClient::overrideMediaType() const
{
    notImplemented();
    return String();
}

void WebFrameLoaderClient::dispatchDidClearWindowObjectInWorld(DOMWrapperWorld& world)
{
    WebCore::Frame* webFrame = core(m_webFrame);
    if ( webFrame == &webFrame->page()->mainFrame() )
    {
        m_webFrame->webView()->RebindJavaScriptObject();
    }

    if ( &world != &mainThreadNormalWorld() )
	{
		return;
	}

#if USE(TSRENDER)
    if ( !m_timer.isActive() )
    {
        m_timer.startOneShot( 0_s );
    }
#endif //#if USE(TSRENDER)

    WebView *webView = m_webFrame->webView();
	WebFrameLoadDelegate *frameLoadDelegate = webView->frameLoadDelegate();
	if ( !frameLoadDelegate )
	{
		return;
	}

	

    frameLoadDelegate->didClearWindowObjectForFrameInScriptWorld(webView, m_webFrame);
}

Ref<FrameNetworkingContext> WebFrameLoaderClient::createNetworkingContext()
{
    return WebFrameNetworkingContext::create(core(m_webFrame));
}

bool WebFrameLoaderClient::shouldAlwaysUsePluginDocument(const String& mimeType) const
{
    return false;
}

void WebFrameLoaderClient::revertToProvisionalState(DocumentLoader*)
{
    notImplemented();
}

void WebFrameLoaderClient::setMainFrameDocumentReady(bool)
{
    notImplemented();
}

void WebFrameLoaderClient::cancelPolicyCheck()
{
}

void WebFrameLoaderClient::prefetchDNS(const String& hostname)
{
    WebCore::prefetchDNS(hostname);
}